/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : BooG - evaluate Capacitance
 * Date    : 2000
 * Author  : Francois Donnet
 */


#include MUT_H
#include MLO_H
#include ABL_H
#include ABE_H
#include "bog_lib_utils.h"
#include "bog_lib_negativ.h"
#include "bog_lib_cell.h"
#include "bog_signal_utils.h"
#include "bog_signal_nameindex.h"
#include "bog_normalize_message.h"
#include "bog_normalize_ARITY.h"
#include "bog_map_prepare.h"
#include "bog_map_adapt.h"


/******************************************************************************/
/* swap the contain of 2 abl                                                  */
/******************************************************************************/
static void swap_pointers(chain_list* abl1, chain_list* abl2)
{
   chain_list* cdr, *car;
   
   car=ABL_CAR(abl1);
   cdr=ABL_CDR(abl1);
   ABL_CAR(abl1)=ABL_CAR(abl2);
   ABL_CDR(abl1)=ABL_CDR(abl2);
   ABL_CAR(abl2)=car;
   ABL_CDR(abl2)=cdr;
}


/******************************************************************************/
/* adapt expr to match cell library                                           */
/* if negativ is non zero opposite value if expr is wanted                    */
/* evaluate capacitance of leaves                                             */
/******************************************************************************/
extern chain_list* loc_adapt_abl(chain_list* expr, float C)
{   
   cell_list* cell;
   port_list* port, *top;
   chain_list* abl;

   if (!expr) {
      fprintf(stderr,"loc_adapt_abl: NULL pointer\n");
      exit(1);
   }

   /*do not need to insert a buffer*/
   if (ABL_ATOM(expr)) {
      inccapacitance(ABL_ATOM_VALUE(expr),C);
      return expr;
   }
   
   cell=cell_prepare(expr);

   /*arity is too big to be matched*/
   if (!cell) {   
      /*evaluate with the biggest oper*/
      int arity=ABL_ARITY(expr);  /*memorize arity*/
      /*search the biggest arity which matches expr*/
      for (ABL_ARITY(expr)=ABL_ARITY(expr)-1 ; ABL_ARITY(expr)>0; 
      ABL_ARITY(expr)--) {
         cell=cell_prepare(expr);
         if (cell) break;
      }
      ABL_ARITY(expr)=arity;   /*put back arity*/
      if (!cell) {
         fprintf(stderr,"Mapping Error: No cell could match  '");
         display_abl(expr);
         if (ABL_ATOM(expr)) fprintf(stderr,"'\n");
         else fprintf(stderr,"'    (oper arity=%d)\n",ABL_ARITY(expr));
         exit(1);      
      }
      
      /*take the first value*/
      for (port=cell->PORT; port; port=port->NEXT) if (port->C!=0) break;
      for (abl=ABL_CDR(expr); abl; abl=ABL_CDR(abl)) {
         ABL_CAR(abl)=loc_adapt_abl(ABL_CAR(abl),port->C);
      }
      return expr;
   }
   
   top=copyport(cell->PORT);  /*because this port can be used later*/
   
   for (port=top; port; port=port->NEXT) {
    if (isvss(port->NAME) || isvdd(port->NAME)) continue;
    else switch (port->DIRECTION) {
    case OUT: case TRISTATE: break;
    default: 
      if (!port->ABL) {
         fprintf(stderr,
         "loc_adapt_abl: compute error %s not really match on port %s\n",
         cell->NAME,port->NAME);
         exit(1);
      }
      if (port->NEGATIV) {
        if (ABL_ATOM(port->ABL)) 
          ABL_ATOM_VALUE(port->ABL)=getoppositename(ABL_ATOM_VALUE(port->ABL));
        else  
        switch (ABL_OPER(port->ABL)) {
        case ABL_AND: ABL_OPER(port->ABL)=ABL_NAND; break;
        case ABL_NAND: ABL_OPER(port->ABL)=ABL_AND; break;
        case ABL_OR: ABL_OPER(port->ABL)=ABL_NOR; break;
        case ABL_NOR: ABL_OPER(port->ABL)=ABL_OR; break;
        case ABL_XOR: ABL_OPER(port->ABL)=ABL_NXOR; break;
        case ABL_NXOR: ABL_OPER(port->ABL)=ABL_XOR; break;
        case ABL_NOT: /*impossible, it should have been accepted by pattern*/
        default:
            fprintf(stderr,
            "adapt_abl: oper %s forbidden at this level\n",
            getabloperuppername(ABL_OPER(port->ABL)));
            exit(1);
        }    
      }
      port->ABL=loc_adapt_abl(port->ABL,port->C);
      if (!port->NEGATIV) continue;
      /*insert a NOT to match perfectly with cell*/
      abl=createabloper(ABL_NOT);
      ABL_ARITY(abl)=1;
      /*swap pointers references*/
      swap_pointers(port->ABL,abl);
      ABL_CDR(port->ABL)=addchain(NULL,abl);
    }  
   }   

   delport(top);
   return expr;
}


/******************************************************************************/
/* adapt expr to match cell library                                           */
/* if negativ is non zero opposite value if expr is wanted                    */
/* evaluate capacitance of leaves                                             */
/******************************************************************************/
extern chain_list* adapt_abl(chain_list* expr)
{   
   cell_list* cell;
   port_list* port, *top;
   chain_list* abl;

   if (!expr) {
      fprintf(stderr,"adapt_abl: NULL pointer\n");
      exit(1);
   }

   cell=cell_prepare(expr);

   /*arity is too big to be matched*/
   if (!cell) {   
      int arity;

      if (ABL_ATOM(expr)) {
         /*no buffer?!!!*/
         fprintf(stderr,"Mapping Error: No cell could match  '");
         display_abl(expr);
         fprintf(stderr,"'\n");
         exit(1);
      }
      
      /*evaluate with the biggest oper*/
      arity=ABL_ARITY(expr);  /*memorize arity*/
      /*search the biggest arity which matches expr*/
      for (ABL_ARITY(expr)=ABL_ARITY(expr)-1 ; ABL_ARITY(expr)>0; 
      ABL_ARITY(expr)--) {
         cell=cell_prepare(expr);
         if (cell) break;
      }
      ABL_ARITY(expr)=arity;   /*put back arity*/
      if (!cell) {
         fprintf(stderr,"Mapping Error: No cell could match  '");
         display_abl(expr);
         fprintf(stderr,"'    (oper arity=%d)\n",ABL_ARITY(expr));
         exit(1);      
      }
      
      /*take the first value*/
      for (port=cell->PORT; port; port=port->NEXT) if (port->C!=0) break;
      for (abl=ABL_CDR(expr); abl; abl=ABL_CDR(abl)) {
         ABL_CAR(abl)=loc_adapt_abl(ABL_CAR(abl),port->C);
      }
      return expr;
   }
   
   
   top=copyport(cell->PORT);  /*because this port can be used later*/
   
   for (port=top; port; port=port->NEXT) {
    if (isvss(port->NAME) || isvdd(port->NAME)) continue;
    else switch (port->DIRECTION) {
    case OUT: case TRISTATE: break;
    default: 
      if (!port->ABL) {
         fprintf(stderr,
         "adapt_abl: compute error %s not really match on port %s\n",
         cell->NAME,port->NAME);
         exit(1);
      }
      if (port->NEGATIV) {
        if (ABL_ATOM(port->ABL)) 
          ABL_ATOM_VALUE(port->ABL)=getoppositename(ABL_ATOM_VALUE(port->ABL));
        else  
        switch (ABL_OPER(port->ABL)) {
        case ABL_AND: ABL_OPER(port->ABL)=ABL_NAND; break;
        case ABL_NAND: ABL_OPER(port->ABL)=ABL_AND; break;
        case ABL_OR: ABL_OPER(port->ABL)=ABL_NOR; break;
        case ABL_NOR: ABL_OPER(port->ABL)=ABL_OR; break;
        case ABL_XOR: ABL_OPER(port->ABL)=ABL_NXOR; break;
        case ABL_NXOR: ABL_OPER(port->ABL)=ABL_XOR; break;
        case ABL_NOT: /*impossible, it should have been accepted by pattern*/
        default:
            fprintf(stderr,
            "adapt_abl: oper %s forbidden at this level\n",
            getabloperuppername(ABL_OPER(port->ABL)));
            exit(1);
        }    
      }
      port->ABL=loc_adapt_abl(port->ABL,port->C);
      if (!port->NEGATIV) continue;
      /*insert a NOT to match perfectly with cell*/
      abl=createabloper(ABL_NOT);
      ABL_ARITY(abl)=1;
      /*swap pointers references*/
      swap_pointers(port->ABL,abl);
      ABL_CDR(port->ABL)=addchain(NULL,abl);
    }  
   }   

   delport(top);
   return expr;
}


/******************************************************************************/
/* adapt biabl to match cell library                                          */
/* evaluate capacitance of leaves                                             */
/******************************************************************************/
extern biabl_list* adapt_bus(biabl_list* biabl)
{   
   cell_list* cell;
   port_list *port, *top;
   chain_list* abl;

   if (!biabl) {
      fprintf(stderr,"adapt_bus: NULL pointer\n");
      exit(1);
   }

   /*check if there is a tristate inverter in library*/
   /* is-it an advantage to use a tristate inverter here? */
   if (is_tristate_inverter_lib() &&
    !ABL_ATOM(biabl->VALABL) && ABL_OPER(biabl->VALABL)!=ABL_NOT) {
      /*put a NOT in front of expression to be mapped by bus inverter*/
      biabl->VALABL=inv_oper(biabl->VALABL,1);
      biabl->VALABL=build_negativ(biabl->VALABL);  
      biabl->VALABL=createablnotexpr(biabl->VALABL);
      /* createablnotexpr() can simplify*/
      if (!ABL_ATOM(biabl->VALABL)) ABL_ARITY(biabl->VALABL)=1;
   }

   cell=cell_prepare_bus(biabl);
   
   top=copyport(cell->PORT);  /*because this port can be used later*/
   
   for (port=top; port; port=port->NEXT) {
    if (isvss(port->NAME) || isvdd(port->NAME)) continue;
    else switch (port->DIRECTION) {
    case OUT: case TRISTATE: break;
    default: 
      if (!port->ABL) {
         fprintf(stderr,
         "adapt_bus: compute error %s not really match on port %s\n",
         cell->NAME,port->NAME);
         exit(1);
      }
      if (port->NEGATIV) {
        if (ABL_ATOM(port->ABL)) 
          ABL_ATOM_VALUE(port->ABL)=getoppositename(ABL_ATOM_VALUE(port->ABL));
        else  
        switch (ABL_OPER(port->ABL)) {
        case ABL_AND: ABL_OPER(port->ABL)=ABL_NAND; break;
        case ABL_NAND: ABL_OPER(port->ABL)=ABL_AND; break;
        case ABL_OR: ABL_OPER(port->ABL)=ABL_NOR; break;
        case ABL_NOR: ABL_OPER(port->ABL)=ABL_OR; break;
        case ABL_XOR: ABL_OPER(port->ABL)=ABL_NXOR; break;
        case ABL_NXOR: ABL_OPER(port->ABL)=ABL_XOR; break;
        case ABL_NOT: /*impossible, it should have been accepted by pattern*/
        default:
            fprintf(stderr,
            "adapt_bus: oper %s forbidden at this level\n",
            getabloperuppername(ABL_OPER(port->ABL)));
            exit(1);
        }    
      }
      port->ABL=loc_adapt_abl(port->ABL,port->C);
      if (!port->NEGATIV) continue;
      /*insert a NOT to match perfectly with cell*/
      abl=createabloper(ABL_NOT);
      ABL_ARITY(abl)=1;
      /*swap pointers references*/
      swap_pointers(port->ABL,abl);
      ABL_CDR(port->ABL)=addchain(NULL,abl);
    }  
   }   

   delport(top);
   return biabl;
}


/******************************************************************************/
/* adapt biabl to match cell library                                          */
/* evaluate capacitance of leaves                                             */
/******************************************************************************/
extern biabl_list* adapt_reg(biabl_list* biabl)
{   
   cell_list* cell;
   port_list *port, *top;
   chain_list* abl;
   ptype_list* ptype;

   if (!biabl) {
      fprintf(stderr,"adapt_reg: NULL pointer\n");
      exit(1);
   }

   /*is it a flip-flop?*/
   ptype=getptype(biabl->USER,ABL_STABLE);

   cell=cell_prepare_reg(biabl);
   
   /* if flip-flop is on rising edge and biabl on falling, we must accord them*/
   if (ptype && ABL_ATOM(cell->CNDABL) && !ABL_ATOM(biabl->CNDABL)
    && ABL_OPER(biabl->CNDABL)==ABL_NOT && ABL_ATOM(ABL_CADR(biabl->CNDABL))) {
      chain_list* abl;
      /*build a local inverted clock signal*/
      abl=createablatom(getoppositename(ABL_ATOM_VALUE(ABL_CADR(biabl->CNDABL))));
      freeablexpr(biabl->CNDABL);
      biabl->CNDABL=abl;
      /*retry with a positiv signal*/
      cell=cell_prepare_reg(biabl);
   } 

   top=copyport(cell->PORT);  /*because this port can be used later*/
   
   for (port=top; port; port=port->NEXT) {
    if (isvss(port->NAME) || isvdd(port->NAME)) continue;
    else switch (port->DIRECTION) {
    case OUT: case TRISTATE: break;
    default: 
      if (!port->ABL) {
         fprintf(stderr,
         "adapt_reg: compute error %s not really match on port %s\n",
         cell->NAME,port->NAME);
         exit(1);
      }
      if (port->NEGATIV) {
        if (ABL_ATOM(port->ABL)) 
          ABL_ATOM_VALUE(port->ABL)=getoppositename(ABL_ATOM_VALUE(port->ABL));
        else  
        switch (ABL_OPER(port->ABL)) {
        case ABL_AND: ABL_OPER(port->ABL)=ABL_NAND; break;
        case ABL_NAND: ABL_OPER(port->ABL)=ABL_AND; break;
        case ABL_OR: ABL_OPER(port->ABL)=ABL_NOR; break;
        case ABL_NOR: ABL_OPER(port->ABL)=ABL_OR; break;
        case ABL_XOR: ABL_OPER(port->ABL)=ABL_NXOR; break;
        case ABL_NXOR: ABL_OPER(port->ABL)=ABL_XOR; break;
        case ABL_NOT: /*impossible, it should have been accepted by pattern*/
        default:
            fprintf(stderr,
            "adapt_reg: oper %s forbidden at this level\n",
            getabloperuppername(ABL_OPER(port->ABL)));
            exit(1);
        }    
      }
      port->ABL=loc_adapt_abl(port->ABL,port->C);
      if (!port->NEGATIV) continue;
      /*insert a NOT to match perfectly with cell*/
      abl=createabloper(ABL_NOT);
      ABL_ARITY(abl)=1;
      /*swap pointers references*/
      swap_pointers(port->ABL,abl);
      ABL_CDR(port->ABL)=addchain(NULL,abl);
    }  
   }   

   delport(top);
   return biabl;
}



