/*
 *  $Id: util_Floorplan.c,v 1.3 2000/09/19 08:00:06 jpc Exp $
 *
 *  /----------------------------------------------------------------\
 *  |                                                                |
 *  |        A l l i a n c e   C A D   S y s t e m                   |
 *  |  S i l i c o n   E n s e m b l e / A l l i a n c e             |
 *  |                                                                |
 *  |  Author    :                      Jean-Paul CHAPUT             |
 *  |  E-mail    :         alliance-support@asim.lip6.fr             |
 *  | ============================================================== |
 *  |  C Module    :         "./util_Floorplan.c"                    |
 *  | ************************************************************** |
 *  |  U p d a t e s                                                 |
 *  |                                                                |
 *  \----------------------------------------------------------------/
 */


# include  "util_Defs.h"


/*  ------------------------------------------------------------------
 *  Local constants.
 */

# define        F_NO_OVERLAP        ((long)(1<<0))
# define        F_OVERLAP           ((long)(1<<1))
# define        F_I1_IN_I2          ((long)(1<<2))
# define        F_I2_IN_I1          ((long)(1<<3))
# define        F_SEG_EXCLUS        ((long)(1<<0))
# define        F_SEG_INCLUS        ((long)(1<<1))
# define        F_SEG_HALFOVER      ((long)(1<<2))
# define        F_SEG_OVER          ((long)(1<<3))


/*  ------------------------------------------------------------------
 *  Local variables.
 */

  static         char  routingLayers3[256] = "l_alu1.l_alu2.l_alu3.";
  static         char  routingLayers6[256] =
                        "l_alu1.l_alu2.l_alu3.l_alu4.l_alu5.l_alu6.";


/*  ----------------------------------------------------------------------
 *  Local functions declarations.
 */

  static   long     computeArea     __FP((struct lofig *apLofig));
  static   long     intervalRel     __FP((long  aI1_min,
                                          long  aI1_max,
                                          long  aI2_min,
                                          long  aI2_max));
  static long       segInArea       __FP((struct phseg *aPhseg,
                                                  long  aX1,
                                                  long  aY1,
                                                  long  aX2,
                                                  long  aY2));


/*
 *  /--------------------------------------------------------------------\
 *  |                     Functions Definitions                          |
 *  \--------------------------------------------------------------------/
 */

/*  ----------------------------------------------------------------------
 *  Function  :  "computeArea()".
 */

static long  computeArea(apLofig)
  struct lofig *apLofig;
{
  struct loins* pLoins;
  struct phfig* pModel;
          long  area;
          long  AB_width, AB_height;


  area = 0L;

  for (pLoins = apLofig->LOINS; pLoins != NULL; pLoins = pLoins->NEXT) {
    pModel = getphfig (pLoins->FIGNAME, 'P');

    AB_width  = pModel->XAB2 - pModel->XAB1;
    AB_height = pModel->YAB2 - pModel->YAB1;


    if (AB_width <= 0) {
      eprinth (NULL);
      eprintf (
        "Negative or null width (%ld) for abutment box of \"%s\".\n",
        AB_width, pModel->NAME);
      EXIT (1);
    }

    if (AB_height <= 0) {
      eprinth (NULL);
      eprintf (
        "Negative or null height (%ld) for abutment box of \"%s\".\n",
        AB_height, pModel->NAME);
      EXIT (1);
    }

    if (AB_width % MBK_X_GRID) {
      eprinth (NULL);
      eprintf (
        "Unpitched width (%ld) for abutment box of \"%s\".\n",
        AB_width, pModel->NAME);
      EXIT (1);
    }

    if (AB_height % MBK_Y_SLICE) {
      eprinth (NULL);
      eprintf (
        "Unpitched height (%ld) for abutment box of \"%s\".\n",
        AB_height, pModel->NAME);
      EXIT (1);
    }


    AB_width  = AB_width  / MBK_X_GRID;
    AB_height = AB_height / MBK_Y_SLICE;

    area += AB_width * AB_height;
  }


  return (area);
}


/*  ----------------------------------------------------------------------
 *  Function  :  "makeFloorplan()".
 */

extern struct phfig *makeFloorplan(apLofig,
                                   aMargin,
                                   aXSize,
                                   aYSize,
                                   aAspectRatio,
                                   aPower,
                                   aFlags)
  struct lofig *apLofig;
          long  aMargin;
          long  aXSize;
          long  aYSize;
        double  aAspectRatio;
          long  aPower;
          long  aFlags;
{
  struct phfig *pPhfig;
          long  specCount, i;
          long  area;
          long  xSize, ySize;
          long  XAB1, YAB1, XAB2, YAB2;
          char  row_name[1024], row_type[1024];
          char *routingLayers;


  specCount  = (aFlags & F_FLOOR_X_SIZE) ? 1 : 0;
  specCount += (aFlags & F_FLOOR_Y_SIZE) ? 1 : 0;
  specCount += (aFlags & F_FLOOR_ASPECT_RATIO) ? 1 : 0;

  if (specCount != 1) {
    errMBK  ("makeFloorplan");
    eprintf ("\n  One, and only one argument among aXSize, aYSize and");
    eprintf (   " aAspectRatio must\n   be non zero.\n");
    EXIT (1);
  }

  pPhfig = addphfig (apLofig->NAME);

  area = computeArea (apLofig);
  area = (long)((1.0 + (double)aMargin / 100.0) * (double)area);

  /* Shut up GCC. */
  xSize = 0;
  ySize = 0;

  if (aFlags & F_FLOOR_X_SIZE) {
    xSize = aXSize;
    ySize =    area / (xSize - 7 * aPower)
           + ((area % (xSize - 7 * aPower)) ? 1 : 0);
  }

  if (aFlags & F_FLOOR_Y_SIZE) {
    ySize = aYSize;
    xSize = (7 * aPower) + (area / aYSize) + ((area / aYSize) ? 1 : 0);
  }

  if (aFlags & F_FLOOR_ASPECT_RATIO) {
    double  x, r, b, xy;

    xy = (double)(Y_SLICE / X_GRID);

    b = ((double)(7 * aPower)) / xy;
    x = (b + sqrt (pow (b, 2.0) + 4 * xy * (double)(area) * aAspectRatio))
      / (2 * xy * aAspectRatio);

    ySize = (long)floor (x * aAspectRatio);

    /* Compute missing area (as ySize is rounded down).
     * The unit the number of horizontal sites.
     */ 
    r = ((double)(area + 7 * aPower * ySize) - pow ((double)ySize, 2.0) * xy)
      /  (double)(ySize);

    xSize = ((long)ceil ((double)ySize * xy + r));
  }


  XAB1 = 0L;
  YAB1 = 0L;
  XAB2 = xSize * MBK_X_GRID;
  YAB2 = ySize * MBK_Y_SLICE;


  /* Build the floorplan. */

  /* The Abutment-box. */
  defab (pPhfig, XAB1, YAB1, XAB2, YAB2);

  /* The row matrix. */
  strcpy (row_type, "core");
  for (i = 0; i < ySize; i++) {
    sprintf (row_name, "row_%ld", i);

    DEF2MBK_row (pPhfig,
                 row_name,
                 row_type,
                 (i % 2) ? DEF_N : DEF_FS,
                 xSize,
                 1L,
                 DEF_X_GRID,
                 DEF_Y_SLICE,
                 XAB1,
                 YAB1 + i * MBK_Y_SLICE);
  }

  /* The tracks matrix. */
  routingLayers = routingLayers6;

  if (aFlags & F_FLOOR_LAYERS_3) {
    routingLayers = routingLayers3;
  }

  /* Horizontal tracks. */
  DEF2MBK_track (pPhfig,
                 "Y",
                 MBK2DEF_length (YAB1 + MBK_X_GRID),
                 ySize * (Y_SLICE / X_GRID) - 1,
                 DEF_X_GRID,
                 routingLayers,
                 XAB1,
                 YAB1 + MBK_X_GRID);

  /* Vertical tracks. */
  DEF2MBK_track (pPhfig,
                 "X",
                 MBK2DEF_length (XAB1 + MBK_X_GRID),
                 xSize - 1,
                 DEF_X_GRID,
                 routingLayers,
                 XAB1 + MBK_X_GRID,
                 YAB1);


  return (pPhfig);
}


/*  ----------------------------------------------------------------------
 *  Function  :  "expandFloorplan()".
 */

extern void  expandFloorplan(apPhfig)
  struct phfig *apPhfig;
{
  apPhfig->XAB1 -= FLOOR_XY_EXPAND;
  apPhfig->YAB1 -= FLOOR_XY_EXPAND;
  apPhfig->XAB2 += FLOOR_XY_EXPAND;
  apPhfig->YAB2 += FLOOR_XY_EXPAND;

  DEF2MBK_blockage (apPhfig,
                    MBK2DEF_length (apPhfig->XAB1),
                    MBK2DEF_length (apPhfig->YAB1),
                    MBK2DEF_length (FLOOR_XY_EXPAND),
                    MBK2DEF_length (FLOOR_XY_EXPAND)
                    );

  DEF2MBK_blockage (apPhfig,
                    MBK2DEF_length (apPhfig->XAB1),
                    MBK2DEF_length (apPhfig->YAB2 - FLOOR_XY_EXPAND),
                    MBK2DEF_length (FLOOR_XY_EXPAND),
                    MBK2DEF_length (FLOOR_XY_EXPAND)
                    );

  DEF2MBK_blockage (apPhfig,
                    MBK2DEF_length (apPhfig->XAB2 - FLOOR_XY_EXPAND),
                    MBK2DEF_length (apPhfig->YAB1),
                    MBK2DEF_length (FLOOR_XY_EXPAND),
                    MBK2DEF_length (FLOOR_XY_EXPAND)
                    );

  DEF2MBK_blockage (apPhfig,
                    MBK2DEF_length (apPhfig->XAB2 - FLOOR_XY_EXPAND),
                    MBK2DEF_length (apPhfig->YAB2 - FLOOR_XY_EXPAND),
                    MBK2DEF_length (FLOOR_XY_EXPAND),
                    MBK2DEF_length (FLOOR_XY_EXPAND)
                    );
}


/*  ----------------------------------------------------------------------
 *  Function  :  "intervalRel()".
 */

static long  intervalRel(aI1_min, aI1_max, aI2_min, aI2_max)
  long  aI1_min, aI1_max, aI2_min, aI2_max;
{
  long  flag;


  if (aI1_min > aI1_max) {
    errMBK  ("intervalRel");
    eprintf (" I1_min(%ld) > I1_max(%ld) !\n", aI1_min, aI1_max);
    EXIT (1);
  }
  if (aI2_min > aI2_max) {
    errMBK  ("intervalRel");
    eprintf (" I2_min(%ld) > I2_max(%ld) !\n", aI2_min, aI2_max);
    EXIT (1);
  }


  flag = 2;

  if ((aI1_min < aI2_min) || (aI1_min > aI2_max)) flag--;
  if ((aI1_max < aI2_min) || (aI1_max > aI2_max)) flag--;
  if ((aI1_min < aI2_min) && (aI1_max > aI2_max)) flag = 3;


  switch (flag) {
    default:
    case 3: flag = F_I2_IN_I1;   break;
    case 2: flag = F_I1_IN_I2;   break;
    case 1: flag = F_OVERLAP;    break;
    case 0: flag = F_NO_OVERLAP; break;
  }

  return (flag);
}


/*  ----------------------------------------------------------------------
 *  Function  :  "segInArea()".
 */

static long  segInArea(aPhseg, aX1, aY1, aX2, aY2)
  struct phseg *aPhseg;
          long  aX1, aY1, aX2, aY2;
{
  long  flagX, flagY;


  if (aX1 > aX2) {
    errMBK  ("segInArea");
    eprintf ("Bad area X1(%ld) > X2(%ld) !\n", aX1, aX2);
    EXIT (1);
  }
  if (aY1 > aY2) {
    errMBK  ("segInArea");
    eprintf ("Bad area Y1(%ld) > Y2(%ld) !\n", aY1, aY2);
    EXIT (1);
  }


  flagX = intervalRel (aPhseg->X1, aPhseg->X2, aX1, aX2);
  flagY = intervalRel (aPhseg->Y1, aPhseg->Y2, aY1, aY2);


  if ((flagX == F_NO_OVERLAP) || (flagY == F_NO_OVERLAP))
    return (F_SEG_EXCLUS);

  if ((flagX == F_I1_IN_I2) && (flagY == F_I1_IN_I2))
    return (F_SEG_INCLUS);

  if ((flagX == F_I2_IN_I1) && (flagY == F_I2_IN_I1))
    return (F_SEG_HALFOVER);

  return (F_SEG_OVER);
}


/*  ----------------------------------------------------------------------
 *  Function  :  "shrinkFloorplan()".
 */

extern void  shrinkFloorplan(apPhfig)
  struct phfig *apPhfig;
{
  struct phseg *pPhseg;
  struct phcon *pPhcon;
  struct phref *pPhref;
          long  XAB1, YAB1, XAB2, YAB2;


  XAB1 = apPhfig->XAB1;
  YAB1 = apPhfig->YAB1;
  XAB2 = apPhfig->XAB2;
  YAB2 = apPhfig->YAB2;

  /* Shrink the abutment-box. */
  apPhfig->XAB1 += FLOOR_XY_EXPAND;
  apPhfig->YAB1 += FLOOR_XY_EXPAND;
  apPhfig->XAB2 -= FLOOR_XY_EXPAND;
  apPhfig->YAB2 -= FLOOR_XY_EXPAND;


  /* Move terminals to the new AB. */
  for (pPhcon = apPhfig->PHCON; pPhcon != NULL; pPhcon = pPhcon->NEXT) {
    if (pPhcon->XCON == XAB1) pPhcon->XCON = apPhfig->XAB1;
    if (pPhcon->XCON == XAB2) pPhcon->XCON = apPhfig->XAB2;
    if (pPhcon->YCON == YAB1) pPhcon->YCON = apPhfig->YAB1;
    if (pPhcon->YCON == YAB2) pPhcon->YCON = apPhfig->YAB2;
  }


  /* Shrink segments to fit in the new AB. */
  for (pPhseg = apPhfig->PHSEG; pPhseg != NULL; pPhseg = pPhseg->NEXT) {
    if (pPhseg->X1 == XAB1) pPhseg->X1 = apPhfig->XAB1;
    if (pPhseg->X2 == XAB2) pPhseg->X2 = apPhfig->XAB2;
    if (pPhseg->Y1 == YAB1) pPhseg->Y1 = apPhfig->YAB1;
    if (pPhseg->Y2 == YAB2) pPhseg->Y2 = apPhfig->YAB2;
  }


  /* Move references to the new AB. */
  for (pPhref = apPhfig->PHREF; pPhref != NULL; pPhref = pPhref->NEXT) {
    if (pPhref->XREF == XAB1) pPhref->XREF = apPhfig->XAB1;
    if (pPhref->XREF == XAB2) pPhref->XREF = apPhfig->XAB2;
    if (pPhref->YREF == YAB1) pPhref->YREF = apPhfig->YAB1;
    if (pPhref->YREF == YAB2) pPhref->YREF = apPhfig->YAB2;
  }
}
